/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Digital Analyzer for Milliwatt Application
 * Mwanalyze 1.0
 * 
 * Copyright (C) 2006, Roger Schreiter
 *
 * Roger Schreiter <roger@planinternet.de>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <stdio.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/lock.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

#ifndef PI
#define PI (4*atan(1))
#endif

static char *tdesc = "Digital Analyzer for Milliwatt Application (slin)";
static char *app = "Mwanalyze";
static char *synopsis = "Spectral analyses based on a given frequency (slin)";
static char *descrip =
  "Mwanalyze(frequency|timeslice|duration|treshold): Analyzes audio stream for tone of given frequency.\n"
  "It computes the amplitude, the average ripple per timeslice, and the number of\n"
  "timeslices with ripple greater then given treshold.\n"
  "frequency is an integer divisible by 4 read as divider of 8 kHz (thus max frequency is 2 kHz).\n"
  "timeslice is an integer read in units of 1/8000 sec.\n"
  "duration is an integer read in units of timeslice.\n"
  "treshold is an integer meaning units of slin.\n"
  "Example: Mwanalyze(8|8000|60|328) will examine a 1000 Hz tone for a minute and\n"
  "reporting e.g. how many seconds had a ripple of more then 1% of slin's fullscale.\n"
  "When application is terminated, read channel variables mwa_ampitude, mwa_ripple, and mwa_bad_timeslices!\n";

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

static int mwanalyze_exec(struct ast_channel *chan, void *data)
{
	int res=0;
	struct localuser *u;
	char *args, *cp, *cp0, linebuf[320];
	unsigned int frequency=0, timeslice=0, duration=0, j;
	int i, v, treshold=0;
	
	int original_read_fmt;
	int original_write_fmt;
	struct ast_frame *inf = NULL;
	struct ast_frame outf;
	unsigned char waste[AST_FRIENDLY_OFFSET];	/* seen in app_millitwatt.c, don't understand what this is useful for */
	short buf[640];
	unsigned int indexp, len, ts_index, ts_offset, phase;
	short sin_x, cos_x, val;
	long long sum_sin, sum_cos, total_deviation=0, sum_sin0=0, sum_cos0=0;
	int total_samples, *periods_in_timeslice, total_bad_timeslices;
	double coeff_sin, coeff_cos, amplitude, timeslice_ripple, total_ripple, modell_val, avg_val;
	short *tone, *lastperiod;
	long long *avg_period;
	        

	/*** check whether arguments are provided ... */
	
	if ( (data==NULL) || (!((char*)data)[0]) )
	{
		ast_log(LOG_WARNING, "Mwanalyze needs arguments frequency, timeslice, duration and treshold.\n");
		return -1;
	}
	
	if ((args = (char*)malloc(strlen((char*)data))) == NULL)
	{
		ast_log(LOG_WARNING, "Mwanalyze cannot allocate enough memory.\n");
		return -1;
	}
	
	
	/*** now parse arguments ... */
	
	strcpy(args, (char*)data);
	cp = args;
	for (i=0; i<4; i++)
	{
		cp0 = cp;
		cp = strchr(cp, '|');
		if (cp)
		{
			*cp++ = 0;
		} else if (i<3) {
			ast_log(LOG_WARNING, "Mwanalyze needs arguments frequency, timeslice, duration and treshold.\n");
			free(args);
			return -1;
		}
		
		if (sscanf(cp0, "%d", &v)!=1)
		{
			ast_log(LOG_WARNING, "Mwanalyze needs only integers as arguments.\n");
			free(args);
			return -1;
		}
		
		switch (i) {
			case 0: frequency = v;
			case 1: timeslice = v;
			case 2: duration = v;
			case 3: treshold = v;
		} 
	}
	
	/*** prepare sine wave */
	if (frequency%4) frequency--;
	if (frequency%4) frequency--;
	if (frequency%4) frequency--;
	if (frequency<4) {
		frequency=4;
	}
	if (timeslice<frequency) {
		timeslice=frequency;
	}
	if (duration<1) {
		duration=1;
	}
	if (treshold<1) {
		treshold=1;
	}
	
	if ((tone = (short*)malloc(frequency*sizeof(short))) == NULL)
	{
		ast_log(LOG_WARNING, "Mwanalyze cannot allocate enough memory.\n");
		free(args);
		return -1;
	}
	if ((lastperiod = (short*)malloc(frequency*sizeof(short))) == NULL)
	{
		ast_log(LOG_WARNING, "Mwanalyze cannot allocate enough memory.\n");
		free(tone);
		free(args);
		return -1;
	}
	if ((avg_period = (long long*)malloc(frequency*sizeof(long long))) == NULL)
	{
		ast_log(LOG_WARNING, "Mwanalyze cannot allocate enough memory.\n");
		free(lastperiod);
		free(tone);
		free(args);
		return -1;
	}
	if ((periods_in_timeslice = (int*)malloc(frequency*sizeof(int))) == NULL)
	{
		ast_log(LOG_WARNING, "Mwanalyze cannot allocate enough memory.\n");
		free(avg_period);
		free(lastperiod);
		free(tone);
		free(args);
		return -1;
	}
	
	for (indexp=0; indexp<frequency; indexp++)
	{
		tone[indexp] = (short)(sin((double)indexp/frequency*2*PI)*((1<<15)-1));
	}
	
	
	LOCAL_USER_ADD(u);


	/*** answer if necessary */
	if ( (!res) && (chan->_state != AST_STATE_UP) )
	{
		if ((res = ast_answer(chan)) < 0)
		{
			ast_log(LOG_WARNING, "Could not answer channel '%s'\n", chan->name);
		}
	}


	/*** save audio codecs and switch to slin if necessary */
	original_read_fmt = AST_FORMAT_SLINEAR;
	if ( (!res) && ((original_read_fmt=chan->readformat) != AST_FORMAT_SLINEAR) )
	{
		if ((res = ast_set_read_format(chan, AST_FORMAT_SLINEAR)) < 0)
		{
			ast_log(LOG_WARNING, "Unable to set to linear read mode, giving up.\n");
		}
	}
	original_write_fmt = AST_FORMAT_SLINEAR;
	if ( (!res) && ((original_write_fmt=chan->writeformat) != AST_FORMAT_SLINEAR) )
	{
		if ((res = ast_set_write_format(chan, AST_FORMAT_SLINEAR)) < 0)
		{
			ast_log(LOG_WARNING, "Unable to set to linear write mode, giving up.\n");
		}
	}


	ts_index=0;
	ts_offset=0;
	phase=0;
	sum_sin = sum_cos = 0.0;
	total_samples = 0;
	total_ripple = 0.0;
	total_bad_timeslices = 0;
	
	while ( (ts_index<duration) && (!res) && (ast_waitfor(chan, -1) > -1) ) {
		if ((inf = ast_read(chan)) == NULL)
		{
			/*** got hangup */
			res = -1;
			break;
		}
		
		if (inf->frametype == AST_FRAME_VOICE)
		{
			len = inf->samples;
			if (len > sizeof(buf))
			{
				ast_log(LOG_WARNING,"Can only exchange %d bytes.\n", (int)sizeof(buf));
				len = sizeof(buf);
			}

			/*** analyze */
			for (i = 0; i < (int)len; i++)
			{
				/*** Fourier */
				val = ((short*)inf->data)[i];
				sin_x = tone[phase];
				cos_x = tone[(phase<3*frequency/4)?(phase+frequency/4):(phase-3*frequency/4)];
				sum_sin += sin_x*val;
				sum_cos += cos_x*val;
				
				/*** total deviatian from real waveshape */
				if (ts_offset==0)
				{
					total_deviation = 0;
					sum_sin0 = sum_sin;
					sum_cos0 = sum_cos;
				}
				if (ts_offset>=frequency)
				{
					total_deviation += (val>=lastperiod[phase])?(val-lastperiod[phase]):(lastperiod[phase]-val);
				} else {
					avg_period[phase] = 0;
					periods_in_timeslice[phase] = 0;
				}
				lastperiod[phase] = val;
				avg_period[phase] += val;
				periods_in_timeslice[phase]++;
				
				
				/*** adavance pointers */
				if (++ts_offset == timeslice)
				{
					timeslice_ripple = (double)total_deviation/2/timeslice;
					coeff_sin = (sum_sin-sum_sin0)*2.0/((1<<15)-1)/timeslice;
					coeff_cos = (sum_cos-sum_cos0)*2.0/((1<<15)-1)/timeslice;
					
					total_deviation = 0;
					for (j=0; j<frequency; j++)
					{
						sin_x = tone[j];
						cos_x = tone[(j<3*frequency/4)?(j+frequency/4):(j-3*frequency/4)];
						modell_val = coeff_sin * sin_x / ((1<<15)-1) + coeff_cos * cos_x / ((1<<15)-1);
						avg_val = (double)avg_period[j] / periods_in_timeslice[j];
						total_deviation += (long long)((avg_val>=modell_val)?(avg_val-modell_val):(modell_val-avg_val));
					}
					timeslice_ripple += (double)total_deviation/frequency;
					total_ripple += timeslice_ripple;
					if (timeslice_ripple>treshold)
					{
						total_bad_timeslices++;
					}
					
					ts_offset = 0;
					ts_index++;
				}
				
				if (++phase == frequency)
				{
					phase = 0;
				}
				total_samples++;	
			}
			
			waste[0] = 0; /* ???? */
			outf.frametype = AST_FRAME_VOICE;
			outf.subclass = AST_FORMAT_SLINEAR;
			outf.offset = AST_FRIENDLY_OFFSET;	/* ???? */
			outf.mallocd = 0;
			outf.data = buf;
			outf.datalen = len*sizeof(short);
			outf.samples = len;
			outf.src = "app_mwanalyze";
			outf.delivery.tv_sec = 0;
			outf.delivery.tv_usec = 0;
			for (i = 0; i < (int)len; i++)
			{
				if (indexp==frequency) {
					indexp = 0;
				}
				buf[i] = tone[indexp++];
			}
			
			if ((res=ast_write(chan,&outf)) < 0)
			{
				ast_log(LOG_WARNING,"Failed to write frame to '%s'.\n",chan->name);
			}
			
		}

		ast_frfree(inf);
	}


	/*** restore audio codecs if necessary */
	if (original_read_fmt != AST_FORMAT_SLINEAR)
	{
		if (ast_set_read_format(chan, original_read_fmt) < 0)
		{
			ast_log(LOG_WARNING, "Unable to restore read mode.\n");
		}
	}
	if (original_write_fmt != AST_FORMAT_SLINEAR)
	{
		if (ast_set_write_format(chan, original_write_fmt) < 0)
		{
			ast_log(LOG_WARNING, "Unable to restore write mode.\n");
		}
	}


	coeff_sin = (total_samples)?(sum_sin*2.0/((1<<15)-1)/total_samples):0.0;
	coeff_cos = (total_samples)?(sum_cos*2.0/((1<<15)-1)/total_samples):0.0;
	amplitude = sqrt(coeff_sin*coeff_sin + coeff_cos*coeff_cos);

	sprintf(linebuf, "%lg", amplitude);
	pbx_builtin_setvar_helper(chan, "mwa_amplitude", linebuf);
	sprintf(linebuf, "%lg", (duration)?(total_ripple/duration):0.0);
	pbx_builtin_setvar_helper(chan, "mwa_ripple", linebuf);
	sprintf(linebuf, "%d", total_bad_timeslices);
	pbx_builtin_setvar_helper(chan, "mwa_bad_timeslices", linebuf);

	LOCAL_USER_REMOVE(u);
	free(periods_in_timeslice);
	free(avg_period);
	free(lastperiod);
	free(tone);
	free(args);
	return res;
}


int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
	return ast_unregister_application(app);
}

int load_module(void)
{
	return ast_register_application(app, mwanalyze_exec, synopsis, tdesc);
}

char *description(void)
{
	return descrip;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
